@@ -1,24 +1,15 @@
-/* ====================================================================
- * Copyright (c) 2007-2007 Open Source Applications Foundation.
+/*
+ *   Licensed under the Apache License, Version 2.0 (the "License");
+ *   you may not use this file except in compliance with the License.
+ *   You may obtain a copy of the License at
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions: 
+ *       http://www.apache.org/licenses/LICENSE-2.0
  *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software. 
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- * ====================================================================
+ *   Unless required by applicable law or agreed to in writing, software
+ *   distributed under the License is distributed on an "AS IS" BASIS,
+ *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ *   See the License for the specific language governing permissions and
+ *   limitations under the License.
  */
 
 #include <jni.h>
@@ -31,6 +22,10 @@
 #include "java/lang/reflect/Method.h"
 #include "java/lang/reflect/Constructor.h"
 #include "java/lang/reflect/Field.h"
+#ifdef _java_generics
+#include "java/lang/reflect/Type.h"
+#include "java/lang/reflect/TypeVariable.h"
+#endif
 
 namespace java {
     namespace lang {
@@ -51,10 +46,17 @@ namespace java {
             mid_isAssignableFrom,
             mid_getComponentType,
             mid_getSuperclass,
+            mid_getDeclaringClass,
+            mid_getEnclosingClass,
             mid_getInterfaces,
             mid_getName,
             mid_getModifiers,
             mid_isInstance,
+#ifdef _java_generics
+            mid_getTypeParameters,
+            mid_getGenericInterfaces,
+            mid_getGenericSuperclass,
+#endif
             max_mid
         };
 
@@ -110,6 +112,12 @@ namespace java {
                 _mids[mid_getSuperclass] =
                     env->getMethodID(cls, "getSuperclass",
                                      "()Ljava/lang/Class;");
+                _mids[mid_getDeclaringClass] =
+                    env->getMethodID(cls, "getDeclaringClass",
+                                     "()Ljava/lang/Class;");
+                _mids[mid_getEnclosingClass] =
+                    env->getMethodID(cls, "getEnclosingClass",
+                                     "()Ljava/lang/Class;");
                 _mids[mid_getInterfaces] =
                     env->getMethodID(cls, "getInterfaces",
                                      "()[Ljava/lang/Class;");
@@ -122,7 +130,17 @@ namespace java {
                 _mids[mid_isInstance] =
                     env->getMethodID(cls, "isInstance",
                                      "(Ljava/lang/Object;)Z");
-
+#ifdef _java_generics
+                _mids[mid_getTypeParameters] =
+                    env->getMethodID(cls, "getTypeParameters",
+                                     "()[Ljava/lang/reflect/TypeVariable;");
+                _mids[mid_getGenericInterfaces] =
+                    env->getMethodID(cls, "getGenericInterfaces",
+                                     "()[Ljava/lang/reflect/Type;");
+                _mids[mid_getGenericSuperclass] =
+                    env->getMethodID(cls, "getGenericSuperclass",
+                                     "()Ljava/lang/reflect/Type;");
+#endif
                 class$ = (Class *) new JObject(cls);
             }
 
@@ -218,6 +236,16 @@ namespace java {
             return Class(env->callObjectMethod(this$, _mids[mid_getSuperclass]));
         }
 
+        Class Class::getDeclaringClass() const
+        {
+            return Class(env->callObjectMethod(this$, _mids[mid_getDeclaringClass]));
+        }
+
+        Class Class::getEnclosingClass() const
+        {
+            return Class(env->callObjectMethod(this$, _mids[mid_getEnclosingClass]));
+        }
+
         JArray<Class> Class::getInterfaces() const
         {
             jobjectArray array = (jobjectArray)
@@ -241,6 +269,23 @@ namespace java {
             return env->callBooleanMethod(this$, _mids[mid_isInstance],
                                           obj.this$);
         }
+
+#ifdef _java_generics
+        JArray<TypeVariable> Class::getTypeParameters() const
+        {
+            return JArray<TypeVariable>(env->callObjectMethod(this$, _mids[mid_getTypeParameters]));
+        }
+
+        JArray<Type> Class::getGenericInterfaces() const
+        {
+            return JArray<Type>(env->callObjectMethod(this$, _mids[mid_getGenericInterfaces]));
+        }
+
+        Type Class::getGenericSuperclass() const
+        {
+            return Type(env->callObjectMethod(this$, _mids[mid_getGenericSuperclass]));
+        }
+#endif
     }
 }
 
@@ -253,6 +298,8 @@ namespace java {
     namespace lang {
         using namespace reflect;
 
+        static PyObject *t_Class_cast_(PyTypeObject *type, PyObject *arg);
+        static PyObject *t_Class_instance_(PyTypeObject *type, PyObject *arg);
         static PyObject *t_Class_forName(PyTypeObject *type, PyObject *arg);
         static PyObject *t_Class_getDeclaredConstructors(t_Class *self);
         static PyObject *t_Class_getDeclaredMethods(t_Class *self);
@@ -267,11 +314,30 @@ namespace java {
         static PyObject *t_Class_isAssignableFrom(t_Class *self, PyObject *arg);
         static PyObject *t_Class_getComponentType(t_Class *self);
         static PyObject *t_Class_getSuperclass(t_Class *self);
+        static PyObject *t_Class_getDeclaringClass(t_Class *self);
+        static PyObject *t_Class_getEnclosingClass(t_Class *self);
         static PyObject *t_Class_getInterfaces(t_Class *self);
         static PyObject *t_Class_getName(t_Class *self);
         static PyObject *t_Class_getModifiers(t_Class *self);
+#ifdef _java_generics
+        static PyObject *t_Class_getTypeParameters(t_Class *self);
+        static PyObject *t_Class_getGenericInterfaces(t_Class *self);
+        static PyObject *t_Class_getGenericSuperclass(t_Class *self);
+        static PyObject *t_Class_get__parameters_(t_Class *self, void *data);
+
+        static PyGetSetDef t_Class__fields_[] = {
+            DECLARE_GET_FIELD(t_Class, parameters_),
+            { NULL, NULL, NULL, NULL, NULL }
+        };
+#else
+        static PyGetSetDef t_Class__fields_[] = {
+            { NULL, NULL, NULL, NULL, NULL }
+        };
+#endif
 
         static PyMethodDef t_Class__methods_[] = {
+            DECLARE_METHOD(t_Class, cast_, METH_O | METH_CLASS),
+            DECLARE_METHOD(t_Class, instance_, METH_O | METH_CLASS),
             DECLARE_METHOD(t_Class, forName, METH_O | METH_CLASS),
             DECLARE_METHOD(t_Class, getDeclaredConstructors, METH_NOARGS),
             DECLARE_METHOD(t_Class, getDeclaredMethods, METH_NOARGS),
@@ -286,14 +352,46 @@ namespace java {
             DECLARE_METHOD(t_Class, isAssignableFrom, METH_O),
             DECLARE_METHOD(t_Class, getComponentType, METH_NOARGS),
             DECLARE_METHOD(t_Class, getSuperclass, METH_NOARGS),
+            DECLARE_METHOD(t_Class, getDeclaringClass, METH_NOARGS),
+            DECLARE_METHOD(t_Class, getEnclosingClass, METH_NOARGS),
             DECLARE_METHOD(t_Class, getInterfaces, METH_NOARGS),
             DECLARE_METHOD(t_Class, getName, METH_NOARGS),
             DECLARE_METHOD(t_Class, getModifiers, METH_NOARGS),
+#ifdef _java_generics
+            DECLARE_METHOD(t_Class, getTypeParameters, METH_NOARGS),
+            DECLARE_METHOD(t_Class, getGenericInterfaces, METH_NOARGS),
+            DECLARE_METHOD(t_Class, getGenericSuperclass, METH_NOARGS),
+#endif
             { NULL, NULL, 0, NULL }
         };
 
         DECLARE_TYPE(Class, t_Class, Object, java::lang::Class,
-                     abstract_init, 0, 0, 0, 0, 0);
+                     abstract_init, 0, 0, t_Class__fields_, 0, 0);
+
+#ifdef _java_generics
+        PyObject *t_Class::wrap_Object(const Class& object, PyTypeObject *T)
+        {
+            PyObject *obj = t_Class::wrap_Object(object);
+            if (obj != Py_None)
+            {
+                t_Class *self = (t_Class *) obj;
+                self->parameters[0] = T;
+            }
+            return obj;
+        }
+#endif
+        static PyObject *t_Class_cast_(PyTypeObject *type, PyObject *arg)
+        {
+            if (!(arg = castCheck(arg, Class::initializeClass, 1)))
+                return NULL;
+            return t_Class::wrap_Object(Class(((t_Class *) arg)->object.this$));
+        }
+        static PyObject *t_Class_instance_(PyTypeObject *type, PyObject *arg)
+        {
+            if (!castCheck(arg, Class::initializeClass, 0))
+                Py_RETURN_FALSE;
+            Py_RETURN_TRUE;
+        }
 
         static PyObject *t_Class_forName(PyTypeObject *type, PyObject *arg)
         {
@@ -307,9 +405,14 @@ namespace java {
                 char *className = PyString_AsString(arg);
                 String name = String(env->fromUTF(className));
 
-                return t_Class::wrapObject(Class::forName(name));
-            } catch (JCCEnv::exception e) {
-                return PyErr_SetJavaError(e.throwable);
+                return t_Class::wrap_Object(Class::forName(name));
+            } catch (int e) {
+                switch (e) {
+                  case _EXC_JAVA:
+                    return PyErr_SetJavaError();
+                  default:
+                    throw;
+                }
             }
         }
 
@@ -318,7 +421,7 @@ namespace java {
             JArray<Constructor> constructors((jobject) NULL);
 
             OBJ_CALL(constructors = self->object.getDeclaredConstructors());
-            return constructors.toSequence(t_Constructor::wrapObject);
+            return constructors.toSequence(t_Constructor::wrap_Object);
         }
 
         static PyObject *t_Class_getDeclaredMethods(t_Class *self)
@@ -326,7 +429,7 @@ namespace java {
             JArray<Method> methods((jobject) NULL);
 
             OBJ_CALL(methods = self->object.getDeclaredMethods());
-            return methods.toSequence(t_Method::wrapObject);
+            return methods.toSequence(t_Method::wrap_Object);
         }
 
         static PyObject *t_Class_getMethods(t_Class *self)
@@ -334,7 +437,7 @@ namespace java {
             JArray<Method> methods((jobject) NULL);
 
             OBJ_CALL(methods = self->object.getMethods());
-            return methods.toSequence(t_Method::wrapObject);
+            return methods.toSequence(t_Method::wrap_Object);
         }
 
         static PyObject *t_Class_getMethod(t_Class *self, PyObject *args)
@@ -346,7 +449,7 @@ namespace java {
             if (!parseArgs(args, "s[j", Class::class$, &name, &params))
             {
                 OBJ_CALL(method = self->object.getMethod(name, params));
-                return t_Method::wrapObject(method);
+                return t_Method::wrap_Object(method);
             }
 
             return PyErr_SetArgsError((PyObject *) self, "getMethod", args);
@@ -361,7 +464,7 @@ namespace java {
             if (!parseArgs(args, "s[j", Class::class$, &name, &params))
             {
                 OBJ_CALL(method = self->object.getDeclaredMethod(name, params));
-                return t_Method::wrapObject(method);
+                return t_Method::wrap_Object(method);
             }
 
             return PyErr_SetArgsError((PyObject *) self, "getMethod", args);
@@ -372,7 +475,7 @@ namespace java {
             JArray<Field> fields((jobject) NULL);
 
             OBJ_CALL(fields = self->object.getDeclaredFields());
-            return fields.toSequence(t_Field::wrapObject);
+            return fields.toSequence(t_Field::wrap_Object);
         }
 
         static PyObject *t_Class_getDeclaredClasses(t_Class *self)
@@ -380,7 +483,7 @@ namespace java {
             JArray<Class> array((jobject) NULL);
 
             OBJ_CALL(array = self->object.getDeclaredClasses());
-            return array.toSequence(t_Class::wrapObject);
+            return array.toSequence(t_Class::wrap_Object);
         }
 
         static PyObject *t_Class_isArray(t_Class *self)
@@ -409,7 +512,7 @@ namespace java {
 
         static PyObject *t_Class_isAssignableFrom(t_Class *self, PyObject *arg)
         {
-            if (!PyObject_TypeCheck(arg, &ClassType))
+            if (!PyObject_TypeCheck(arg, &PY_TYPE(Class)))
             {
                 PyErr_SetObject(PyExc_TypeError, arg);
                 return NULL;
@@ -420,8 +523,13 @@ namespace java {
                 int isAssignableFrom = self->object.isAssignableFrom(cls);
 
                 Py_RETURN_BOOL(isAssignableFrom);
-            } catch (JCCEnv::exception e) {
-                return PyErr_SetJavaError(e.throwable);
+            } catch (int e) {
+                switch (e) {
+                  case _EXC_JAVA:
+                    return PyErr_SetJavaError();
+                  default:
+                    throw;
+                }
             }
         }
 
@@ -430,7 +538,7 @@ namespace java {
             Class cls((jobject) NULL);
 
             OBJ_CALL(cls = self->object.getComponentType());
-            return t_Class::wrapObject(cls);
+            return t_Class::wrap_Object(cls);
         }
 
         static PyObject *t_Class_getSuperclass(t_Class *self)
@@ -438,7 +546,23 @@ namespace java {
             Class cls((jobject) NULL);
 
             OBJ_CALL(cls = self->object.getSuperclass());
-            return t_Class::wrapObject(cls);
+            return t_Class::wrap_Object(cls);
+        }
+
+        static PyObject *t_Class_getDeclaringClass(t_Class *self)
+        {
+            Class cls((jobject) NULL);
+
+            OBJ_CALL(cls = self->object.getDeclaringClass());
+            return t_Class::wrap_Object(cls);
+        }
+
+        static PyObject *t_Class_getEnclosingClass(t_Class *self)
+        {
+            Class cls((jobject) NULL);
+
+            OBJ_CALL(cls = self->object.getEnclosingClass());
+            return t_Class::wrap_Object(cls);
         }
 
         static PyObject *t_Class_getInterfaces(t_Class *self)
@@ -446,7 +570,7 @@ namespace java {
             JArray<Class> interfaces((jobject) NULL);
 
             OBJ_CALL(interfaces = self->object.getInterfaces());
-            return interfaces.toSequence(t_Class::wrapObject);
+            return interfaces.toSequence(t_Class::wrap_Object);
         }
 
         static PyObject *t_Class_getName(t_Class *self)
@@ -465,5 +589,35 @@ namespace java {
             return PyInt_FromLong(modifiers);            
         }
 
+#ifdef _java_generics
+        static PyObject *t_Class_getTypeParameters(t_Class *self)
+        {
+            JArray<TypeVariable> result((jobject) NULL);
+            OBJ_CALL(result = self->object.getTypeParameters());
+
+            return result.toSequence(t_TypeVariable::wrap_Object);
+        }
+
+        static PyObject *t_Class_getGenericInterfaces(t_Class *self)
+        {
+            JArray<Type> result((jobject) NULL);
+            OBJ_CALL(result = self->object.getGenericInterfaces());
+
+            return result.toSequence(t_Type::wrap_Object);
+        }
+
+        static PyObject *t_Class_getGenericSuperclass(t_Class *self)
+        {
+            Type result((jobject) NULL);
+            OBJ_CALL(result = self->object.getGenericSuperclass());
+
+            return t_Type::wrap_Object(result);
+        }
+
+        static PyObject *t_Class_get__parameters_(t_Class *self, void *data)
+        {
+            return typeParameters(self->parameters, sizeof(self->parameters));
+        }
+#endif
     }
 }
