@@ -1,24 +1,15 @@
-/* ====================================================================
- * Copyright (c) 2007-2008 Open Source Applications Foundation.
+/*
+ *   Licensed under the Apache License, Version 2.0 (the "License");
+ *   you may not use this file except in compliance with the License.
+ *   You may obtain a copy of the License at
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions: 
+ *       http://www.apache.org/licenses/LICENSE-2.0
  *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software. 
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
- * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- * ====================================================================
+ *   Unless required by applicable law or agreed to in writing, software
+ *   distributed under the License is distributed on an "AS IS" BASIS,
+ *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ *   See the License for the specific language governing permissions and
+ *   limitations under the License.
  */
 
 #include <map>
@@ -27,8 +18,6 @@
 
 #include "JCCEnv.h"
 
-extern JCCEnv *env;
-
 #if defined(_MSC_VER) || defined(__WIN32)
 _DLL_EXPORT DWORD VM_ENV = 0;
 #else
@@ -81,13 +70,25 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
     }
 #endif
 
+    if (vm)
+        set_vm(vm, vm_env);
+    else
+        this->vm = NULL;
+}
+
+void JCCEnv::set_vm(JavaVM *vm, JNIEnv *vm_env)
+{
     this->vm = vm;
     set_vm_env(vm_env);
-    env = this;
 
     _sys = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/System"));
     _obj = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/Object"));
+#ifdef _jcc_lib
+    _thr = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("org/apache/jcc/PythonException"));
+#else
     _thr = (jclass) vm_env->NewGlobalRef(vm_env->FindClass("java/lang/RuntimeException"));
+#endif
+
     _mids = new jmethodID[max_mid];
 
     _mids[mid_sys_identityHashCode] =
@@ -96,6 +97,9 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
     _mids[mid_sys_setProperty] =
         vm_env->GetStaticMethodID(_sys, "setProperty",
                                   "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;");
+    _mids[mid_sys_getProperty] =
+        vm_env->GetStaticMethodID(_sys, "getProperty",
+                                  "(Ljava/lang/String;)Ljava/lang/String;");
     _mids[mid_obj_toString] =
         vm_env->GetMethodID(_obj, "toString",
                             "()Ljava/lang/String;");
@@ -105,9 +109,73 @@ JCCEnv::JCCEnv(JavaVM *vm, JNIEnv *vm_env)
     _mids[mid_obj_getClass] =
         vm_env->GetMethodID(_obj, "getClass",
                             "()Ljava/lang/Class;");
-    _mids[mid_thr_getMessage] =
-        vm_env->GetMethodID(_thr, "getMessage",
-                            "()Ljava/lang/String;");
+
+
+    jclass iterable = vm_env->FindClass("java/lang/Iterable");
+
+    if (iterable == NULL) /* JDK < 1.5 */
+    {
+        vm_env->ExceptionClear();
+        _mids[mid_iterator] = NULL;
+        _mids[mid_iterator_next] = NULL;
+    }
+    else
+    {
+        _mids[mid_iterator] =
+            vm_env->GetMethodID(iterable,
+                                "iterator", "()Ljava/util/Iterator;");
+        _mids[mid_iterator_next] =
+            vm_env->GetMethodID(vm_env->FindClass("java/util/Iterator"),
+                                "next", "()Ljava/lang/Object;");
+    }
+
+
+    _mids[mid_enumeration_nextElement] =
+        vm_env->GetMethodID(vm_env->FindClass("java/util/Enumeration"),
+                            "nextElement", "()Ljava/lang/Object;");
+
+    _mids[mid_Boolean_booleanValue] =
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Boolean"),
+                            "booleanValue", "()Z");
+    _mids[mid_Byte_byteValue] = 
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Byte"),
+                            "byteValue", "()B");
+    _mids[mid_Character_charValue] =
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Character"),
+                            "charValue", "()C");
+    _mids[mid_Double_doubleValue] = 
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Double"),
+                            "doubleValue", "()D");
+    _mids[mid_Float_floatValue] =
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Float"),
+                            "floatValue", "()F");
+    _mids[mid_Integer_intValue] = 
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Integer"),
+                            "intValue", "()I");
+    _mids[mid_Long_longValue] = 
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Long"),
+                            "longValue", "()J");
+    _mids[mid_Short_shortValue] = 
+        vm_env->GetMethodID(vm_env->FindClass("java/lang/Short"),
+                            "shortValue", "()S");
+}
+
+int JCCEnv::attachCurrentThread(char *name, int asDaemon)
+{
+    JNIEnv *jenv = NULL;
+    JavaVMAttachArgs attach = {
+        JNI_VERSION_1_4, name, NULL
+    };
+    int result;
+
+    if (asDaemon)
+        result = vm->AttachCurrentThreadAsDaemon((void **) &jenv, &attach);
+    else
+        result = vm->AttachCurrentThread((void **) &jenv, &attach);
+
+    set_vm_env(jenv);
+
+    return result;
 }
 
 #if defined(_MSC_VER) || defined(__WIN32)
@@ -130,19 +198,80 @@ void JCCEnv::set_vm_env(JNIEnv *vm_env)
 
 #endif
 
-jclass JCCEnv::findClass(const char *className)
+jint JCCEnv::getJNIVersion() const
+{
+    return get_vm_env()->GetVersion();
+}
+
+jstring JCCEnv::getJavaVersion() const
+{
+    return (jstring)
+        callStaticObjectMethod(_sys, _mids[mid_sys_getProperty],
+                               get_vm_env()->NewStringUTF("java.version"));
+}
+
+jobject JCCEnv::iterator(jobject obj) const
+{
+    return callObjectMethod(obj, _mids[mid_iterator]);
+}
+
+jobject JCCEnv::iteratorNext(jobject obj) const
+{
+    return callObjectMethod(obj, _mids[mid_iterator_next]);
+}
+
+jobject JCCEnv::enumerationNext(jobject obj) const
+{
+    return callObjectMethod(obj, _mids[mid_enumeration_nextElement]);
+}
+
+jboolean JCCEnv::isInstanceOf(jobject obj, getclassfn initializeClass) const
+{
+    return get_vm_env()->IsInstanceOf(obj, (*initializeClass)());
+}
+
+jclass JCCEnv::findClass(const char *className) const
 {
     jclass cls = NULL;
 
-    if (env)
-        cls = get_vm_env()->FindClass(className);
+    if (vm)
+    {
+        JNIEnv *vm_env = get_vm_env();
+
+        if (vm_env)
+            cls = vm_env->FindClass(className);
+#ifdef PYTHON
+        else
+        {
+            PythonGIL gil;
+
+            PyErr_SetString(PyExc_RuntimeError, "attachCurrentThread() must be called first");
+            throw _EXC_PYTHON;
+        }
+#else
+        else
+            throw _EXC_JAVA;
+#endif
+    }
+#ifdef PYTHON
+    else
+    {
+        PythonGIL gil;
+
+        PyErr_SetString(PyExc_RuntimeError, "initVM() must be called first");
+        throw _EXC_PYTHON;
+    }
+#else
+    else
+        throw _EXC_JAVA;
+#endif
 
     reportException();
 
     return cls;
 }
 
-void JCCEnv::registerNatives(jclass cls, JNINativeMethod *methods, int n)
+void JCCEnv::registerNatives(jclass cls, JNINativeMethod *methods, int n) const
 {
     get_vm_env()->RegisterNatives(cls, methods, n);
 }
@@ -162,13 +291,19 @@ jobject JCCEnv::newGlobalRef(jobject obj, int id)
                     break;
                 if (isSame(obj, iter->second.global))
                 {
+                    /* If it's in the table but not the same reference,
+                     * it must be a local reference and must be deleted.
+                     */
+                    if (obj != iter->second.global)
+                        get_vm_env()->DeleteLocalRef(obj);
+                        
                     iter->second.count += 1;
                     return iter->second.global;
                 }
             }
 
-            countedRef ref;
             JNIEnv *vm_env = get_vm_env();
+            countedRef ref;
 
             ref.global = vm_env->NewGlobalRef(obj);
             ref.count = 1;
@@ -178,7 +313,7 @@ jobject JCCEnv::newGlobalRef(jobject obj, int id)
             return ref.global;
         }
         else
-            return get_vm_env()->NewWeakGlobalRef(obj);
+            return (jobject) get_vm_env()->NewWeakGlobalRef(obj);
     }
 
     return NULL;
@@ -201,11 +336,25 @@ jobject JCCEnv::deleteGlobalRef(jobject obj, int id)
                 {
                     if (iter->second.count == 1)
                     {
-                        get_vm_env()->DeleteGlobalRef(iter->second.global);
+                        JNIEnv *vm_env = get_vm_env();
+
+                        if (!vm_env)
+                        {
+                            /* Python's cyclic garbage collector may remove
+                             * an object inside a thread that is not attached
+                             * to the JVM. This makes sure the JVM doesn't
+                             * segfault.
+                             */
+                            attachCurrentThread(NULL, 0);
+                            vm_env = get_vm_env();
+                        }
+
+                        vm_env->DeleteGlobalRef(iter->second.global);
                         refs.erase(iter);
                     }
                     else
                         iter->second.count -= 1;
+
                     return NULL;
                 }
             }
@@ -213,23 +362,39 @@ jobject JCCEnv::deleteGlobalRef(jobject obj, int id)
             printf("deleting non-existent ref: 0x%x\n", id);
         }
         else
-            get_vm_env()->DeleteWeakGlobalRef(obj);
+            get_vm_env()->DeleteWeakGlobalRef((jweak) obj);
     }
 
     return NULL;
 }
 
-jobject JCCEnv::newObject(jclass (*initializeClass)(), jmethodID **mids,
+jobject JCCEnv::newObject(getclassfn initializeClass, jmethodID **mids,
                           int m, ...)
 {
-    jclass cls;
+    jclass cls = (*initializeClass)();
+    JNIEnv *vm_env = get_vm_env();
     jobject obj;
-    va_list ap;
 
-    va_start(ap, m);
-    cls = (*initializeClass)();
-    obj = get_vm_env()->NewObjectV(cls, (*mids)[m], ap);
-    va_end(ap);
+    if (vm_env)
+    {
+        va_list ap;
+
+        va_start(ap, m);
+        obj = vm_env->NewObjectV(cls, (*mids)[m], ap);
+        va_end(ap);
+    }
+#ifdef PYTHON
+    else
+    {
+        PythonGIL gil;
+
+        PyErr_SetString(PyExc_RuntimeError, "attachCurrentThread() must be called first");
+        throw _EXC_PYTHON;
+    }
+#else
+    else
+        throw _EXC_JAVA;
+#endif
 
     reportException();
 
@@ -244,13 +409,14 @@ jobjectArray JCCEnv::newObjectArray(jclass cls, int size)
     return array;
 }
 
-void JCCEnv::setObjectArrayElement(jobjectArray array, int n, jobject obj)
+void JCCEnv::setObjectArrayElement(jobjectArray array, int n,
+                                   jobject obj) const
 {
     get_vm_env()->SetObjectArrayElement(array, n, obj);
     reportException();
 }
 
-jobject JCCEnv::getObjectArrayElement(jobjectArray array, int n)
+jobject JCCEnv::getObjectArrayElement(jobjectArray array, int n) const
 {
     jobject obj = get_vm_env()->GetObjectArrayElement(array, n);
 
@@ -258,7 +424,7 @@ jobject JCCEnv::getObjectArrayElement(jobjectArray array, int n)
     return obj;
 }
 
-int JCCEnv::getArrayLength(jarray array)
+int JCCEnv::getArrayLength(jarray array) const
 {
     int len = get_vm_env()->GetArrayLength(array);
 
@@ -266,67 +432,90 @@ int JCCEnv::getArrayLength(jarray array)
     return len;
 }
 
+#ifdef PYTHON
+jclass JCCEnv::getPythonExceptionClass() const
+{
+    return _thr;
+}
+#endif
 
-void JCCEnv::reportException()
+void JCCEnv::reportException() const
 {
     JNIEnv *vm_env = get_vm_env();
     jthrowable throwable = vm_env->ExceptionOccurred();
 
     if (throwable)
     {
-        //vm_env->ExceptionDescribe();
-        vm_env->ExceptionClear();
+        if (!env->handlers)
+            vm_env->ExceptionDescribe();
 
 #ifdef PYTHON
-        /* kludge: if RuntimeException is thrown assume a Python error */
-        jobject cls =
-            vm_env->CallObjectMethod(throwable, _mids[mid_obj_getClass]);
+        PythonGIL gil;
 
-        if (vm_env->IsSameObject(cls, _thr))
-            throw pythonError(throwable);
+        if (PyErr_Occurred())
+        {
+            /* _thr is PythonException ifdef _jcc_lib (shared mode)
+             * if not shared mode, _thr is RuntimeException
+             */
+            jobject cls = (jobject) vm_env->GetObjectClass(throwable);
+
+            if (vm_env->IsSameObject(cls, _thr))
+            {
+#ifndef _jcc_lib
+                /* PythonException class is not available without shared mode.
+                 * Python exception information thus gets lost and exception
+                 * is reported via plain Java RuntimeException.
+                 */
+                PyErr_Clear();
+                throw _EXC_JAVA;
+#else
+                throw _EXC_PYTHON;
+#endif
+            }
+        }
 #endif
 
-        throw exception(throwable);
+        throw _EXC_JAVA;
     }
 }
 
 
-#define DEFINE_CALL(jtype, Type)                                        \
-    jtype JCCEnv::call##Type##Method(jobject obj,                       \
-                                     jmethodID mid, ...)                \
-    {                                                                   \
-        va_list ap;                                                     \
-        jtype result;                                                   \
-                                                                        \
-        va_start(ap, mid);                                              \
-        result = get_vm_env()->Call##Type##MethodV(obj, mid, ap);       \
-        va_end(ap);                                                     \
-                                                                        \
-        reportException();                                              \
-                                                                        \
-        return result;                                                  \
+#define DEFINE_CALL(jtype, Type)                                         \
+    jtype JCCEnv::call##Type##Method(jobject obj,                        \
+                                     jmethodID mid, ...) const           \
+    {                                                                    \
+        va_list ap;                                                      \
+        jtype result;                                                    \
+                                                                         \
+        va_start(ap, mid);                                               \
+        result = get_vm_env()->Call##Type##MethodV(obj, mid, ap);        \
+        va_end(ap);                                                      \
+                                                                         \
+        reportException();                                               \
+                                                                         \
+        return result;                                                   \
     }
 
-#define DEFINE_NONVIRTUAL_CALL(jtype, Type)                             \
-    jtype JCCEnv::callNonvirtual##Type##Method(jobject obj, jclass cls, \
-                                               jmethodID mid, ...)      \
-    {                                                                   \
-        va_list ap;                                                     \
-        jtype result;                                                   \
-                                                                        \
-        va_start(ap, mid);                                              \
-        result = get_vm_env()->CallNonvirtual##Type##MethodV(obj, cls,  \
-                                                             mid, ap);  \
-        va_end(ap);                                                     \
-                                                                        \
-        reportException();                                              \
-                                                                        \
-        return result;                                                  \
+#define DEFINE_NONVIRTUAL_CALL(jtype, Type)                              \
+    jtype JCCEnv::callNonvirtual##Type##Method(jobject obj, jclass cls,  \
+                                               jmethodID mid, ...) const \
+    {                                                                    \
+        va_list ap;                                                      \
+        jtype result;                                                    \
+                                                                         \
+        va_start(ap, mid);                                               \
+        result = get_vm_env()->CallNonvirtual##Type##MethodV(obj, cls,   \
+                                                             mid, ap);   \
+        va_end(ap);                                                      \
+                                                                         \
+        reportException();                                               \
+                                                                         \
+        return result;                                                   \
     }
 
 #define DEFINE_STATIC_CALL(jtype, Type)                                 \
     jtype JCCEnv::callStatic##Type##Method(jclass cls,                  \
-                                           jmethodID mid, ...)          \
+                                           jmethodID mid, ...) const    \
     {                                                                   \
         va_list ap;                                                     \
         jtype result;                                                   \
@@ -370,7 +559,7 @@ DEFINE_STATIC_CALL(jint, Int)
 DEFINE_STATIC_CALL(jlong, Long)
 DEFINE_STATIC_CALL(jshort, Short)
 
-void JCCEnv::callVoidMethod(jobject obj, jmethodID mid, ...)
+void JCCEnv::callVoidMethod(jobject obj, jmethodID mid, ...) const
 {
     va_list ap;
 
@@ -382,7 +571,7 @@ void JCCEnv::callVoidMethod(jobject obj, jmethodID mid, ...)
 }
 
 void JCCEnv::callNonvirtualVoidMethod(jobject obj, jclass cls,
-                                      jmethodID mid, ...)
+                                      jmethodID mid, ...) const
 {
     va_list ap;
 
@@ -393,7 +582,7 @@ void JCCEnv::callNonvirtualVoidMethod(jobject obj, jclass cls,
     reportException();
 }
 
-void JCCEnv::callStaticVoidMethod(jclass cls, jmethodID mid, ...)
+void JCCEnv::callStaticVoidMethod(jclass cls, jmethodID mid, ...) const
 {
     va_list ap;
 
@@ -405,8 +594,49 @@ void JCCEnv::callStaticVoidMethod(jclass cls, jmethodID mid, ...)
 }
 
 
+jboolean JCCEnv::booleanValue(jobject obj) const
+{
+    return get_vm_env()->CallBooleanMethod(obj, _mids[mid_Boolean_booleanValue]);
+}
+
+jbyte JCCEnv::byteValue(jobject obj) const
+{
+    return get_vm_env()->CallByteMethod(obj, _mids[mid_Byte_byteValue]);
+}
+
+jchar JCCEnv::charValue(jobject obj) const
+{
+    return get_vm_env()->CallCharMethod(obj, _mids[mid_Character_charValue]);
+}
+
+jdouble JCCEnv::doubleValue(jobject obj) const
+{
+    return get_vm_env()->CallDoubleMethod(obj, _mids[mid_Double_doubleValue]);
+}
+
+jfloat JCCEnv::floatValue(jobject obj) const
+{
+    return get_vm_env()->CallFloatMethod(obj, _mids[mid_Float_floatValue]);
+}
+
+jint JCCEnv::intValue(jobject obj) const
+{
+    return get_vm_env()->CallIntMethod(obj, _mids[mid_Integer_intValue]);
+}
+
+jlong JCCEnv::longValue(jobject obj) const
+{
+    return get_vm_env()->CallLongMethod(obj, _mids[mid_Long_longValue]);
+}
+
+jshort JCCEnv::shortValue(jobject obj) const
+{
+    return get_vm_env()->CallShortMethod(obj, _mids[mid_Short_shortValue]);
+}
+
+
 jmethodID JCCEnv::getMethodID(jclass cls, const char *name,
-                              const char *signature)
+                              const char *signature) const
 {
     jmethodID id = get_vm_env()->GetMethodID(cls, name, signature);
 
@@ -416,7 +646,7 @@ jmethodID JCCEnv::getMethodID(jclass cls, const char *name,
 }
 
 jfieldID JCCEnv::getFieldID(jclass cls, const char *name,
-                            const char *signature)
+                            const char *signature) const
 {
     jfieldID id = get_vm_env()->GetFieldID(cls, name, signature);
 
@@ -427,7 +657,7 @@ jfieldID JCCEnv::getFieldID(jclass cls, const char *name,
 
 
 jmethodID JCCEnv::getStaticMethodID(jclass cls, const char *name,
-                                    const char *signature)
+                                    const char *signature) const
 {
     jmethodID id = get_vm_env()->GetStaticMethodID(cls, name, signature);
 
@@ -437,7 +667,7 @@ jmethodID JCCEnv::getStaticMethodID(jclass cls, const char *name,
 }
 
 jobject JCCEnv::getStaticObjectField(jclass cls, const char *name,
-                                     const char *signature)
+                                     const char *signature) const
 {
     JNIEnv *vm_env = get_vm_env();
     jfieldID id = vm_env->GetStaticFieldID(cls, name, signature);
@@ -448,7 +678,8 @@ jobject JCCEnv::getStaticObjectField(jclass cls, const char *name,
 }
 
 #define DEFINE_GET_STATIC_FIELD(jtype, Type, signature)                 \
-    jtype JCCEnv::getStatic##Type##Field(jclass cls, const char *name)  \
+    jtype JCCEnv::getStatic##Type##Field(jclass cls,                    \
+                                         const char *name) const        \
     {                                                                   \
         JNIEnv *vm_env = get_vm_env();                                  \
         jfieldID id = vm_env->GetStaticFieldID(cls, name, #signature);  \
@@ -466,7 +697,7 @@ DEFINE_GET_STATIC_FIELD(jlong, Long, J)
 DEFINE_GET_STATIC_FIELD(jshort, Short, S)
 
 #define DEFINE_GET_FIELD(jtype, Type)                                   \
-    jtype JCCEnv::get##Type##Field(jobject obj, jfieldID id)            \
+    jtype JCCEnv::get##Type##Field(jobject obj, jfieldID id) const      \
     {                                                                   \
         jtype value = get_vm_env()->Get##Type##Field(obj, id);          \
         reportException();                                              \
@@ -483,11 +714,12 @@ DEFINE_GET_FIELD(jint, Int)
 DEFINE_GET_FIELD(jlong, Long)
 DEFINE_GET_FIELD(jshort, Short)
 
-#define DEFINE_SET_FIELD(jtype, Type)                                    \
-    void JCCEnv::set##Type##Field(jobject obj, jfieldID id, jtype value) \
-    {                                                                    \
-        get_vm_env()->Set##Type##Field(obj, id, value);                  \
-        reportException();                                               \
+#define DEFINE_SET_FIELD(jtype, Type)                                   \
+    void JCCEnv::set##Type##Field(jobject obj, jfieldID id,             \
+                                  jtype value) const                    \
+    {                                                                   \
+        get_vm_env()->Set##Type##Field(obj, id, value);                 \
+        reportException();                                              \
     }
 
 DEFINE_SET_FIELD(jobject, Object)
@@ -511,12 +743,13 @@ void JCCEnv::setClassPath(const char *classPath)
     jmethodID mf = vm_env->GetMethodID(_fil, "<init>", "(Ljava/lang/String;)V");
     jmethodID mu = vm_env->GetMethodID(_fil, "toURL", "()Ljava/net/URL;");
     jmethodID ma = vm_env->GetMethodID(_ucl, "addURL", "(Ljava/net/URL;)V");
-#ifdef WINDOWS
+#if defined(_MSC_VER) || defined(__WIN32)
     char *pathsep = ";";
+    char *path = _strdup(classPath);
 #else
     char *pathsep = ":";
-#endif
     char *path = strdup(classPath);
+#endif
 
     for (char *cp = strtok(path, pathsep);
          cp != NULL;
@@ -530,7 +763,53 @@ void JCCEnv::setClassPath(const char *classPath)
     free(path);
 }
 
-jstring JCCEnv::fromUTF(const char *bytes)
+char *JCCEnv::getClassPath()
+{
+    JNIEnv *vm_env = get_vm_env();
+    jclass _ucl = (jclass) vm_env->FindClass("java/net/URLClassLoader");
+    jclass _url = (jclass) vm_env->FindClass("java/net/URL");
+    jmethodID mid = vm_env->GetStaticMethodID(_ucl, "getSystemClassLoader",
+                                              "()Ljava/lang/ClassLoader;");
+    jobject classLoader = vm_env->CallStaticObjectMethod(_ucl, mid);
+    jmethodID gu = vm_env->GetMethodID(_ucl, "getURLs", "()[Ljava/net/URL;");
+    jmethodID gp = vm_env->GetMethodID(_url, "getPath", "()Ljava/lang/String;");
+#if defined(_MSC_VER) || defined(__WIN32)
+    char *pathsep = ";";
+#else
+    char *pathsep = ":";
+#endif
+    jobjectArray array = (jobjectArray)
+        vm_env->CallObjectMethod(classLoader, gu);
+    int count = array ? vm_env->GetArrayLength(array) : 0;
+    int first = 1, total = 0;
+    char *classpath = NULL;
+    
+    for (int i = 0; i < count; i++) {
+        jobject url = vm_env->GetObjectArrayElement(array, i);
+        jstring path = (jstring) vm_env->CallObjectMethod(url, gp);
+        const char *chars = vm_env->GetStringUTFChars(path, NULL);
+        int size = vm_env->GetStringUTFLength(path);
+
+        total += size + 1;
+        if (classpath == NULL)
+            classpath = (char *) calloc(total, 1);
+        else
+            classpath = (char *) realloc(classpath, total);
+        if (classpath == NULL)
+            return NULL;
+
+        if (first)
+            first = 0;
+        else
+            strcat(classpath, pathsep);
+
+        strcat(classpath, chars);
+    }
+
+    return classpath;
+}
+
+jstring JCCEnv::fromUTF(const char *bytes) const
 {
     jstring str = get_vm_env()->NewStringUTF(bytes);
 
@@ -539,7 +818,7 @@ jstring JCCEnv::fromUTF(const char *bytes)
     return str;
 }
 
-char *JCCEnv::toUTF(jstring str)
+char *JCCEnv::toUTF(jstring str) const
 {
     JNIEnv *vm_env = get_vm_env();
     int len = vm_env->GetStringUTFLength(str);
@@ -558,14 +837,31 @@ char *JCCEnv::toUTF(jstring str)
     return bytes;
 }
 
-char *JCCEnv::toString(jobject obj)
+char *JCCEnv::toString(jobject obj) const
 {
-    return obj
-        ? toUTF((jstring) callObjectMethod(obj, _mids[mid_obj_toString]))
-        : NULL;
+    try {
+        return obj
+            ? toUTF((jstring) callObjectMethod(obj, _mids[mid_obj_toString]))
+            : NULL;
+    } catch (int e) {
+        switch (e) {
+          case _EXC_PYTHON:
+            return NULL;
+          case _EXC_JAVA: {
+              JNIEnv *vm_env = get_vm_env();
+
+              vm_env->ExceptionDescribe();
+              vm_env->ExceptionClear();
+
+              return NULL;
+          }
+          default:
+            throw;
+        }
+    }
 }
 
-char *JCCEnv::getClassName(jobject obj)
+char *JCCEnv::getClassName(jobject obj) const
 {
     return obj
         ? toString(callObjectMethod(obj, _mids[mid_obj_getClass]))
@@ -574,11 +870,12 @@ char *JCCEnv::getClassName(jobject obj)
 
 #ifdef PYTHON
 
-jstring JCCEnv::fromPyString(PyObject *object)
+jstring JCCEnv::fromPyString(PyObject *object) const
 {
     if (object == Py_None)
         return NULL;
-    else if (PyUnicode_Check(object))
+
+    if (PyUnicode_Check(object))
     {
         if (sizeof(Py_UNICODE) == sizeof(jchar))
         {
@@ -616,43 +913,48 @@ jstring JCCEnv::fromPyString(PyObject *object)
     }
 }
 
-PyObject *JCCEnv::fromJString(jstring js)
+PyObject *JCCEnv::fromJString(jstring js, int delete_local_ref) const
 {
     if (!js)
         Py_RETURN_NONE;
 
     JNIEnv *vm_env = get_vm_env();
+    PyObject *string;
 
     if (sizeof(Py_UNICODE) == sizeof(jchar))
     {
         jboolean isCopy;
         const jchar *buf = vm_env->GetStringChars(js, &isCopy);
         jsize len = vm_env->GetStringLength(js);
-        PyObject *string = PyUnicode_FromUnicode((const Py_UNICODE *) buf, len);
 
+        string = PyUnicode_FromUnicode((const Py_UNICODE *) buf, len);
         vm_env->ReleaseStringChars(js, buf);
-        return string;
     }
     else
     {
         jsize len = vm_env->GetStringLength(js);
-        PyObject *string = PyUnicode_FromUnicode(NULL, len);
 
-        if (!string)
-            return NULL;
+        string = PyUnicode_FromUnicode(NULL, len);
+        if (string)
+        {
+            jboolean isCopy;
+            const jchar *jchars = vm_env->GetStringChars(js, &isCopy);
+            Py_UNICODE *pchars = PyUnicode_AS_UNICODE(string);
 
-        jboolean isCopy;
-        const jchar *jchars = vm_env->GetStringChars(js, &isCopy);
-        Py_UNICODE *pchars = PyUnicode_AS_UNICODE(string);
-
-        for (int i = 0; i < len; i++)
-            pchars[i] = (Py_UNICODE) jchars[i];
+            for (int i = 0; i < len; i++)
+                pchars[i] = (Py_UNICODE) jchars[i];
         
-        vm_env->ReleaseStringChars(js, jchars);
-        return string;
+            vm_env->ReleaseStringChars(js, jchars);
+        }
     }
+
+    if (delete_local_ref)
+        vm_env->DeleteLocalRef((jobject) js);
+
+    return string;
 }
 
+
 /* may be called from finalizer thread which has no vm_env thread local */
 void JCCEnv::finalizeObject(JNIEnv *jenv, PyObject *obj)
 {
